@@ -14,6 +14,7 @@
  * limitations under the License.
  */
 
+#include "ir/load-utils.h"
 #include "ir/utils.h"
 #include "support/hash.h"
 
@@ -91,7 +92,7 @@ Expression* flexibleCopy(Expression* original, Module& wasm, CustomCopier custom
         return builder.makeAtomicLoad(curr->bytes, curr->offset,
                                       copy(curr->ptr), curr->type);
       }
-      return builder.makeLoad(curr->bytes, curr->signed_, curr->offset, curr->align, copy(curr->ptr), curr->type);
+      return builder.makeLoad(curr->bytes, LoadUtils::isSignRelevant(curr) ? curr->signed_ : false, curr->offset, curr->align, copy(curr->ptr), curr->type);
     }
     Expression* visitStore(Store *curr) {
       if (curr->isAtomic) {
@@ -111,12 +112,39 @@ Expression* flexibleCopy(Expression* original, Module& wasm, CustomCopier custom
     Expression* visitAtomicWait(AtomicWait* curr) {
       return builder.makeAtomicWait(copy(curr->ptr), copy(curr->expected), copy(curr->timeout), curr->expectedType, curr->offset);
     }
-    Expression* visitAtomicWake(AtomicWake* curr) {
-      return builder.makeAtomicWake(copy(curr->ptr), copy(curr->wakeCount), curr->offset);
+    Expression* visitAtomicNotify(AtomicNotify* curr) {
+      return builder.makeAtomicNotify(copy(curr->ptr), copy(curr->notifyCount), curr->offset);
     }
-    Expression* visitConst(Const *curr) {
+    Expression* visitSIMDExtract(SIMDExtract* curr) {
+      return builder.makeSIMDExtract(curr->op, copy(curr->vec), curr->index);
+    }
+    Expression* visitSIMDReplace(SIMDReplace* curr) {
+      return builder.makeSIMDReplace(curr->op, copy(curr->vec), curr->index, copy(curr->value));
+    }
+    Expression* visitSIMDShuffle(SIMDShuffle* curr) {
+      return builder.makeSIMDShuffle(copy(curr->left), copy(curr->right), curr->mask);
+    }
+    Expression* visitSIMDBitselect(SIMDBitselect* curr) {
+      return builder.makeSIMDBitselect(copy(curr->left), copy(curr->right), copy(curr->cond));
+    }
+    Expression* visitSIMDShift(SIMDShift* curr) {
+      return builder.makeSIMDShift(curr->op, copy(curr->vec), copy(curr->shift));
+    }
+    Expression* visitConst(Const* curr) {
       return builder.makeConst(curr->value);
     }
+    Expression* visitMemoryInit(MemoryInit* curr) {
+      return builder.makeMemoryInit(curr->segment, copy(curr->dest), copy(curr->offset), copy(curr->size));
+    }
+    Expression* visitDataDrop(DataDrop* curr) {
+      return builder.makeDataDrop(curr->segment);
+    }
+    Expression* visitMemoryCopy(MemoryCopy* curr) {
+      return builder.makeMemoryCopy(copy(curr->dest), copy(curr->source), copy(curr->size));
+    }
+    Expression* visitMemoryFill(MemoryFill* curr) {
+      return builder.makeMemoryFill(copy(curr->dest), copy(curr->value), copy(curr->size));
+    }
     Expression* visitUnary(Unary *curr) {
       return builder.makeUnary(curr->op, copy(curr->value));
     }
