@@ -21,7 +21,18 @@
 #include <cstdint>
 #include <limits>
 
-template <typename T>
+std::vector<char> wasm::read_stdin(Flags::DebugOption debug) {
+  if (debug == Flags::Debug) std::cerr << "Loading stdin..." << std::endl;
+  std::vector<char> input;
+  char c;
+  while (std::cin.get(c) && !std::cin.eof()) {
+    input.push_back(c);
+  }
+  return input;
+}
+
+
+template<typename T>
 T wasm::read_file(const std::string& filename, Flags::BinaryOption binary, Flags::DebugOption debug) {
   if (debug == Flags::Debug) std::cerr << "Loading '" << filename << "'..." << std::endl;
   std::ifstream infile;
@@ -57,6 +68,9 @@ template std::vector<char> wasm::read_file<>(const std::string& , Flags::BinaryO
 
 wasm::Output::Output(const std::string& filename, Flags::BinaryOption binary, Flags::DebugOption debug)
     : outfile(), out([this, filename, binary, debug]() {
+        if (filename == "-") {
+          return std::cout.rdbuf();
+        }
         std::streambuf *buffer;
         if (filename.size()) {
           if (debug == Flags::Debug) std::cerr << "Opening '" << filename << "'" << std::endl;
@@ -84,4 +98,3 @@ size_t wasm::file_size(std::string filename) {
   std::ifstream infile(filename, std::ifstream::ate | std::ifstream::binary);
   return infile.tellg();
 }
-
