@@ -14,7 +14,7 @@
  * limitations under the License.
  */
 
-#include "support/command-line.h"
+#include "tool-options.h"
 
 //
 // Shared optimization options for commandline tools
@@ -22,14 +22,12 @@
 
 namespace wasm {
 
-struct OptimizationOptions : public Options {
+struct OptimizationOptions : public ToolOptions {
   static constexpr const char* DEFAULT_OPT_PASSES = "O";
 
   std::vector<std::string> passes;
-  PassOptions passOptions;
-  FeatureSet features = Feature::Atomics;
 
-  OptimizationOptions(const std::string& command, const std::string& description) : Options(command, description) {
+  OptimizationOptions(const std::string& command, const std::string& description) : ToolOptions(command, description) {
     (*this).add("", "-O", "execute default optimization passes",
                 Options::Arguments::Zero,
                 [this](Options*, const std::string&) {
@@ -94,15 +92,26 @@ struct OptimizationOptions : public Options {
                 [this](Options* o, const std::string& argument) {
                   passOptions.shrinkLevel = atoi(argument.c_str());
                 })
-           .add("--no-validation", "-n", "Disables validation, assumes inputs are correct",
-                Options::Arguments::Zero,
-                [this](Options* o, const std::string& argument) {
-                  passOptions.validate = false;
-                })
            .add("--ignore-implicit-traps", "-iit", "Optimize under the helpful assumption that no surprising traps occur (from load, div/mod, etc.)",
                 Options::Arguments::Zero,
                 [this](Options*, const std::string&) {
                   passOptions.ignoreImplicitTraps = true;
+                })
+           .add("--low-memory-unused", "-lmu", "Optimize under the helpful assumption that the low 1K of memory is not used by the application",
+                Options::Arguments::Zero,
+                [this](Options*, const std::string&) {
+                  passOptions.lowMemoryUnused = true;
+                })
+           .add("--pass-arg", "-pa", "An argument passed along to optimization passes being run. Must be in the form KEY:VALUE",
+                Options::Arguments::N,
+                [this](Options*, const std::string& argument) {
+                  auto colon = argument.find(':');
+                  if (colon == std::string::npos) {
+                    Fatal() << "--pass-arg value must be in the form of KEY:VALUE";
+                  }
+                  auto key = argument.substr(0, colon);
+                  auto value = argument.substr(colon + 1);
+                  passOptions.arguments[key] = value;
                 });
     // add passes in registry
     for (const auto& p : PassRegistry::get()->getRegisteredNames()) {
@@ -132,7 +141,6 @@ struct OptimizationOptions : public Options {
   void runPasses(Module& wasm) {
     PassRunner passRunner(&wasm, passOptions);
     if (debug) passRunner.setDebug(true);
-    passRunner.setFeatures(features);
     for (auto& pass : passes) {
       if (pass == DEFAULT_OPT_PASSES) {
         passRunner.addDefaultOptimizationPasses();
