@@ -29,6 +29,7 @@
 
 #include "pass.h"
 #include "support/command-line.h"
+#include "support/colors.h"
 #include "support/file.h"
 #include "support/path.h"
 #include "support/timing.h"
@@ -50,7 +51,7 @@ std::string GetLastErrorStdStr() {
   if (error) {
     LPVOID lpMsgBuf;
     DWORD bufLen = FormatMessage(
-        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
+        FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,
@@ -78,7 +79,7 @@ struct ProgramResult {
   std::string output;
   double time;
 
-  ProgramResult() {}
+  ProgramResult() = default;
   ProgramResult(std::string command) {
     getFromExecution(command);
   }
@@ -220,11 +221,12 @@ static std::unordered_set<Name> functionsWeTriedToRemove;
 
 struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<Reducer>>> {
   std::string command, test, working;
-  bool verbose, debugInfo;
+  bool binary, deNan, verbose, debugInfo;
 
   // test is the file we write to that the command will operate on
   // working is the current temporary state, the reduction so far
-  Reducer(std::string command, std::string test, std::string working, bool verbose, bool debugInfo) : command(command), test(test), working(working), verbose(verbose), debugInfo(debugInfo) {}
+  Reducer(std::string command, std::string test, std::string working, bool binary, bool deNan, bool verbose, bool debugInfo) :
+    command(command), test(test), working(working), binary(binary), deNan(deNan), verbose(verbose), debugInfo(debugInfo) {}
 
   // runs passes in order to reduce, until we can't reduce any more
   // the criterion here is wasm binary size
@@ -258,6 +260,7 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
       "--reorder-functions",
       "--reorder-locals",
       "--simplify-locals --vacuum",
+      "--strip",
       "--vacuum"
     };
     auto oldSize = file_size(working);
@@ -268,7 +271,9 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
       // try both combining with a generic shrink (so minor pass overhead is compensated for), and without
       for (auto pass : passes) {
         std::string currCommand = Path::getBinaryenBinaryTool("wasm-opt") + " ";
-        currCommand += working + " -o " + test + " " + pass;
+        // TODO(tlively): -all should be replaced with an option to use the
+        // existing feature set, once implemented.
+        currCommand += working + " -all -o " + test + " " + pass;
         if (debugInfo) currCommand += " -g ";
         if (verbose) std::cerr << "|    trying pass command: " << currCommand << "\n";
         if (!ProgramResult(currCommand).failed()) {
@@ -318,6 +323,7 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
     Module wasm;
     ModuleReader reader;
     reader.read(working, *module);
+    wasm.features = FeatureSet::All;
     builder = make_unique<Builder>(*module);
     setModule(module.get());
   }
@@ -340,7 +346,7 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
   bool writeAndTestReduction(ProgramResult& out) {
     // write the module out
     ModuleWriter writer;
-    writer.setBinary(true);
+    writer.setBinary(binary);
     writer.setDebugInfo(debugInfo);
     writer.write(*getModule(), test);
     // note that it is ok for the destructively-reduced module to be bigger
@@ -358,8 +364,22 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
     return (counter % factor) <= bonus;
   }
 
+  bool isOkReplacement(Expression* with) {
+    if (deNan) {
+      if (auto* c = with->dynCast<Const>()) {
+        if (c->value.isNaN()) {
+          return false;
+        }
+      }
+    }
+    return true;
+  }
+
   // tests a reduction on the current traversal node, and undos if it failed
   bool tryToReplaceCurrent(Expression* with) {
+    if (!isOkReplacement(with)) {
+      return false;
+    }
     auto* curr = getCurrent();
     //std::cerr << "try " << curr << " => " << with << '\n';
     if (curr->type != with->type) return false;
@@ -381,6 +401,9 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
 
   // tests a reduction on an arbitrary child
   bool tryToReplaceChild(Expression*& child, Expression* with) {
+    if (!isOkReplacement(with)) {
+      return false;
+    }
     if (child->type != with->type) return false;
     if (!shouldTryToReduce()) return false;
     auto* before = child;
@@ -429,6 +452,25 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
       handleCondition(select->condition);
     } else if (auto* sw = curr->dynCast<Switch>()) {
       handleCondition(sw->condition);
+      // Try to replace switch targets with the default
+      for (auto& target : sw->targets) {
+        if (target != sw->default_) {
+          auto old = target;
+          target = sw->default_;
+          if (!tryToReplaceCurrent(curr)) {
+            target = old;
+          }
+        }
+      }
+      // Try to shorten the list of targets.
+      while (sw->targets.size() > 1) {
+        auto last = sw->targets.back();
+        sw->targets.pop_back();
+        if (!tryToReplaceCurrent(curr)) {
+          sw->targets.push_back(last);
+          break;
+        }
+      }
     } else if (auto* block = curr->dynCast<Block>()) {
       if (!shouldTryToReduce()) return;
       // replace a singleton
@@ -445,13 +487,13 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
           for (Index j = i; j < list.size() - 1; j++) {
             list[j] = list[j + 1];
           }
-          list.resize(list.size() - 1);
+          list.pop_back();
           if (writeAndTestReduction()) {
             std::cerr << "|      block-nop removed\n";
             noteReduction();
             return;
           }
-          list.resize(list.size() + 1);
+          list.push_back(nullptr);
           // we failed; undo
           for (Index j = list.size() - 1; j > i; j--) {
             list[j] = list[j - 1];
@@ -469,7 +511,11 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
     }
     // Finally, try to replace with a child.
     for (auto* child : ChildIterator(curr)) {
-      if (tryToReplaceCurrent(child)) return;
+      if (isConcreteType(child->type) && curr->type == none) {
+        if (tryToReplaceCurrent(builder->makeDrop(child))) return;
+      } else {
+        if (tryToReplaceCurrent(child)) return;
+      }
     }
     // If that didn't work, try to replace with a child + a unary conversion
     if (isConcreteType(curr->type) &&
@@ -477,23 +523,29 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
       for (auto* child : ChildIterator(curr)) {
         if (child->type == curr->type) continue; // already tried
         if (!isConcreteType(child->type)) continue; // no conversion
-        Expression* fixed;
+        Expression* fixed = nullptr;
         switch (curr->type) {
           case i32: {
             switch (child->type) {
+              case i32: WASM_UNREACHABLE();
               case i64: fixed = builder->makeUnary(WrapInt64, child); break;
               case f32: fixed = builder->makeUnary(TruncSFloat32ToInt32, child); break;
               case f64: fixed = builder->makeUnary(TruncSFloat64ToInt32, child); break;
-              default: WASM_UNREACHABLE();
+              case v128: continue; // v128 not implemented yet
+              case none:
+              case unreachable: WASM_UNREACHABLE();
             }
             break;
           }
           case i64: {
             switch (child->type) {
               case i32: fixed = builder->makeUnary(ExtendSInt32, child); break;
+              case i64: WASM_UNREACHABLE();
               case f32: fixed = builder->makeUnary(TruncSFloat32ToInt64, child); break;
               case f64: fixed = builder->makeUnary(TruncSFloat64ToInt64, child); break;
-              default: WASM_UNREACHABLE();
+              case v128: continue; // v128 not implemented yet
+              case none:
+              case unreachable: WASM_UNREACHABLE();
             }
             break;
           }
@@ -501,8 +553,11 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
             switch (child->type) {
               case i32: fixed = builder->makeUnary(ConvertSInt32ToFloat32, child); break;
               case i64: fixed = builder->makeUnary(ConvertSInt64ToFloat32, child); break;
+              case f32: WASM_UNREACHABLE();
               case f64: fixed = builder->makeUnary(DemoteFloat64, child); break;
-              default: WASM_UNREACHABLE();
+              case v128: continue; // v128 not implemented yet
+              case none:
+              case unreachable: WASM_UNREACHABLE();
             }
             break;
           }
@@ -511,11 +566,16 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
               case i32: fixed = builder->makeUnary(ConvertSInt32ToFloat64, child); break;
               case i64: fixed = builder->makeUnary(ConvertSInt64ToFloat64, child); break;
               case f32: fixed = builder->makeUnary(PromoteFloat32, child); break;
-              default: WASM_UNREACHABLE();
+              case f64: WASM_UNREACHABLE();
+              case v128: continue; // v128 not implemented yet
+              case none:
+              case unreachable: WASM_UNREACHABLE();
             }
             break;
           }
-          default: WASM_UNREACHABLE();
+          case v128: continue; // v128 not implemented yet
+          case none:
+          case unreachable: WASM_UNREACHABLE();
         }
         assert(fixed->type == curr->type);
         if (tryToReplaceCurrent(fixed)) return;
@@ -762,7 +822,7 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
     FunctionReferenceRemover referenceRemover(names);
     referenceRemover.walkModule(module.get());
 
-    if (WasmValidator().validate(*module, Feature::All, WasmValidator::Globally | WasmValidator::Quiet) &&
+    if (WasmValidator().validate(*module, WasmValidator::Globally | WasmValidator::Quiet) &&
         writeAndTestReduction()) {
       std::cerr << "|      removed " << names.size() << " functions\n";
       return true;
@@ -804,7 +864,7 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
     // try to replace with a trivial value
     Const* c = builder->makeConst(Literal(int32_t(0)));
     if (tryToReplaceCurrent(c)) return true;
-    c->value = LiteralUtils::makeLiteralFromInt32(1, curr->type);
+    c->value = Literal::makeFromInt32(1, curr->type);
     c->type = curr->type;
     return tryToReplaceCurrent(c);
   }
@@ -829,7 +889,9 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
 
 int main(int argc, const char* argv[]) {
   std::string input, test, working, command;
-  bool verbose = false,
+  bool binary = true,
+       deNan = false,
+       verbose = false,
        debugInfo = false,
        force = false;
   Options options("wasm-reduce", "Reduce a wasm file to a smaller one that has the same behavior on a given command");
@@ -858,6 +920,16 @@ int main(int argc, const char* argv[]) {
              // Add separator just in case
              Path::setBinaryenBinDir(argument + Path::getPathSeparator());
            })
+      .add("--text", "-S", "Emit intermediate files as text, instead of binary (also make sure the test and working files have a .wat or .wast suffix)",
+           Options::Arguments::Zero,
+           [&](Options* o, const std::string& argument) {
+             binary = false;
+           })
+      .add("--denan", "", "Avoid nans when reducing",
+           Options::Arguments::Zero,
+           [&](Options* o, const std::string& argument) {
+             deNan = true;
+           })
       .add("--verbose", "-v", "Verbose output mode",
            Options::Arguments::Zero,
            [&](Options* o, const std::string& argument) {
@@ -888,6 +960,10 @@ int main(int argc, const char* argv[]) {
   if (test.size() == 0) Fatal() << "test file not provided\n";
   if (working.size() == 0) Fatal() << "working file not provided\n";
 
+  if (!binary) {
+    Colors::disable();
+  }
+
   std::cerr << "|wasm-reduce\n";
   std::cerr << "|input: " << input << '\n';
   std::cerr << "|test: " << test << '\n';
@@ -926,7 +1002,11 @@ int main(int argc, const char* argv[]) {
   std::cerr << "|checking that command has expected behavior on canonicalized (read-written) binary\n";
   {
     // read and write it
-    ProgramResult readWrite(Path::getBinaryenBinaryTool("wasm-opt") + " " + input + " -o " + test);
+    // TODO(tlively): -all should be replaced with an option to use the existing
+    // feature set, once implemented.
+    auto cmd = Path::getBinaryenBinaryTool("wasm-opt") + " " + input + " -all -o " + test;
+    if (!binary) cmd += " -S";
+    ProgramResult readWrite(cmd);
     if (readWrite.failed()) {
       stopIfNotForced("failed to read and write the binary", readWrite);
     } else {
@@ -950,7 +1030,7 @@ int main(int argc, const char* argv[]) {
   bool stopping = false;
 
   while (1) {
-    Reducer reducer(command, test, working, verbose, debugInfo);
+    Reducer reducer(command, test, working, binary, deNan, verbose, debugInfo);
 
     // run binaryen optimization passes to reduce. passes are fast to run
     // and can often reduce large amounts of code efficiently, as opposed
@@ -1017,4 +1097,3 @@ int main(int argc, const char* argv[]) {
   std::cerr << "|finished, final size: " << file_size(working) << "\n";
   copy_file(working, test); // just to avoid confusion
 }
-
