@@ -236,11 +236,11 @@ int main(int argc, const char* argv[]) {
   Options options("wasm-shell", "Execute .wast files");
   options
       .add(
-          "--entry", "-e", "call the entry point after parsing the module",
+          "--entry", "-e", "Call the entry point after parsing the module",
           Options::Arguments::One,
           [&entry](Options*, const std::string& argument) { entry = argument; })
       .add(
-          "--skip", "-s", "skip input on certain lines (comma-separated-list)",
+          "--skip", "-s", "Skip input on certain lines (comma-separated-list)",
           Options::Arguments::One,
           [&skipped](Options*, const std::string& argument) {
             size_t i = 0;
@@ -292,6 +292,7 @@ int main(int argc, const char* argv[]) {
         builders[moduleName].swap(builder);
         modules[moduleName].swap(module);
         i++;
+        modules[moduleName]->features = FeatureSet::All;
         bool valid = WasmValidator().validate(*modules[moduleName]);
         if (!valid) {
           WasmPrinter::printModule(modules[moduleName].get());
