@@ -25,6 +25,7 @@
 #include "wasm-traversal.h"
 #include "wasm.h"
 #include "ir/function-type-utils.h"
+#include "ir/import-utils.h"
 #include "ir/module-utils.h"
 
 namespace wasm {
@@ -32,7 +33,12 @@ namespace wasm {
 cashew::IString EMSCRIPTEN_ASM_CONST("emscripten_asm_const");
 cashew::IString EM_JS_PREFIX("__em_js__");
 
-static constexpr const char* dummyFunction = "__wasm_nullptr";
+static Name STACK_SAVE("stackSave"),
+            STACK_RESTORE("stackRestore"),
+            STACK_ALLOC("stackAlloc"),
+            STACK_INIT("stack$init"),
+            POST_INSTANTIATE("__post_instantiate"),
+            ASSIGN_GOT_ENTIRES("__assign_got_enties");
 
 void addExportedFunction(Module& wasm, Function* function) {
   wasm.addFunction(function);
@@ -42,9 +48,31 @@ void addExportedFunction(Module& wasm, Function* function) {
   wasm.addExport(export_);
 }
 
+// TODO(sbc): There should probably be a better way to do this.
+bool isExported(Module& wasm, Name name) {
+  for (auto& ex : wasm.exports) {
+    if (ex->value == name) {
+      return true;
+    }
+  }
+  return false;
+}
+
 Global* EmscriptenGlueGenerator::getStackPointerGlobal() {
-  // Assumption: first global is __stack_pointer
-  return wasm.globals[0].get();
+  // Assumption: The stack pointer is either imported as __stack_pointer or
+  // its the first non-imported and non-exported global.
+  // TODO(sbc): Find a better way to discover the stack pointer.  Perhaps the
+  // linker could export it by name?
+  for (auto& g : wasm.globals) {
+    if (g->imported()) {
+      if (g->base == "__stack_pointer") {
+        return g.get();
+      }
+    } else if (!isExported(wasm, g->name)) {
+      return g.get();
+    }
+  }
+  return nullptr;
 }
 
 Expression* EmscriptenGlueGenerator::generateLoadStackPointer() {
@@ -59,6 +87,8 @@ Expression* EmscriptenGlueGenerator::generateLoadStackPointer() {
     );
   }
   Global* stackPointer = getStackPointerGlobal();
+  if (!stackPointer)
+    Fatal() << "stack pointer global not found";
   return builder.makeGetGlobal(stackPointer->name, i32);
 }
 
@@ -74,14 +104,15 @@ Expression* EmscriptenGlueGenerator::generateStoreStackPointer(Expression* value
     );
   }
   Global* stackPointer = getStackPointerGlobal();
+  if (!stackPointer)
+    Fatal() << "stack pointer global not found";
   return builder.makeSetGlobal(stackPointer->name, value);
 }
 
 void EmscriptenGlueGenerator::generateStackSaveFunction() {
-  Name name("stackSave");
   std::vector<NameType> params { };
   Function* function = builder.makeFunction(
-    name, std::move(params), i32, {}
+    STACK_SAVE, std::move(params), i32, {}
   );
 
   function->body = generateLoadStackPointer();
@@ -90,10 +121,9 @@ void EmscriptenGlueGenerator::generateStackSaveFunction() {
 }
 
 void EmscriptenGlueGenerator::generateStackAllocFunction() {
-  Name name("stackAlloc");
   std::vector<NameType> params { { "0", i32 } };
   Function* function = builder.makeFunction(
-    name, std::move(params), i32, { { "1", i32 } }
+    STACK_ALLOC, std::move(params), i32, { { "1", i32 } }
   );
   Expression* loadStack = generateLoadStackPointer();
   GetLocal* getSizeArg = builder.makeGetLocal(0, i32);
@@ -116,10 +146,9 @@ void EmscriptenGlueGenerator::generateStackAllocFunction() {
 }
 
 void EmscriptenGlueGenerator::generateStackRestoreFunction() {
-  Name name("stackRestore");
   std::vector<NameType> params { { "0", i32 } };
   Function* function = builder.makeFunction(
-    name, std::move(params), none, {}
+    STACK_RESTORE, std::move(params), none, {}
   );
   GetLocal* getArg = builder.makeGetLocal(0, i32);
   Expression* store = generateStoreStackPointer(getArg);
@@ -135,6 +164,136 @@ void EmscriptenGlueGenerator::generateRuntimeFunctions() {
   generateStackRestoreFunction();
 }
 
+static Function* ensureFunctionImport(Module* module, Name name, std::string sig) {
+  // Then see if its already imported
+  ImportInfo info(*module);
+  if (Function* f = info.getImportedFunction(ENV, name)) {
+    return f;
+  }
+  // Failing that create a new function import.
+  auto import = new Function;
+  import->name = name;
+  import->module = ENV;
+  import->base = name;
+  auto* functionType = ensureFunctionType(sig, module);
+  import->type = functionType->name;
+  FunctionTypeUtils::fillFunction(import, functionType);
+  module->addFunction(import);
+  return import;
+}
+
+// Convert LLVM PIC ABI to emscripten ABI
+//
+// When generating -fPIC code llvm will generate imports call GOT.mem and
+// GOT.func in order to access the addresses of external global data and
+// functions.
+//
+// However emscripten uses a different ABI where function and data addresses
+// are available at runtime via special `g$foo` and `fp$bar` function calls.
+//
+// Here we internalize all such wasm globals and generte code that sets their
+// value based on the result of call `g$foo` and `fp$bar` functions at runtime.
+Function* EmscriptenGlueGenerator::generateAssignGOTEntriesFunction() {
+  std::vector<Global*> got_entries_func;
+  std::vector<Global*> got_entries_mem;
+  for (auto& g : wasm.globals) {
+    if (!g->imported()) {
+      continue;
+    }
+    if (g->module == "GOT.func") {
+      got_entries_func.push_back(g.get());
+    } else if (g->module == "GOT.mem") {
+      got_entries_mem.push_back(g.get());
+    } else {
+      continue;
+    }
+    // Make this an internal, non-imported, global.
+    g->module.clear();
+    g->init = Builder(wasm).makeConst(Literal(0));
+  }
+
+  if (!got_entries_func.size() && !got_entries_mem.size()) {
+    return nullptr;
+  }
+
+  Function* assign_func =
+    builder.makeFunction(ASSIGN_GOT_ENTIRES, std::vector<NameType>{}, none, {});
+  Block* block = builder.makeBlock();
+  assign_func->body = block;
+
+  for (Global* g : got_entries_mem) {
+    Name getter(std::string("g$") + g->base.c_str());
+    ensureFunctionImport(&wasm, getter, "i");
+    Expression* call = builder.makeCall(getter, {}, i32);
+    SetGlobal* set_global = builder.makeSetGlobal(g->name, call);
+    block->list.push_back(set_global);
+  }
+
+  for (Global* g : got_entries_func) {
+    Function* f = nullptr;
+    // The function has to exist either as export or an import.
+    // Note that we don't search for the function by name since its internal
+    // name may be different.
+    auto* ex = wasm.getExportOrNull(g->base);
+    if (ex) {
+      assert(ex->kind == ExternalKind::Function);
+      f = wasm.getFunction(ex->value);
+    } else {
+      ImportInfo info(wasm);
+      f = info.getImportedFunction(ENV, g->base);
+    }
+
+    Name getter((std::string("fp$") + g->base.c_str() + std::string("$") + getSig(f)).c_str());
+    ensureFunctionImport(&wasm, getter, "i");
+    Expression* call = builder.makeCall(getter, {}, i32);
+    SetGlobal* set_global = builder.makeSetGlobal(g->name, call);
+    block->list.push_back(set_global);
+  }
+
+  wasm.addFunction(assign_func);
+  return assign_func;
+}
+
+// For emscripten SIDE_MODULE we generate a single exported function called
+// __post_instantiate which calls two functions:
+//
+// - __assign_got_enties
+// - __wasm_call_ctors
+//
+// The former is function we generate here which calls imported g$XXX functions
+// order to assign values to any globals imported from GOT.func or GOT.mem.
+// These globals hold address of functions and globals respectively.
+//
+// The later is the constructor function generaed by lld which performs any
+// fixups on the memory section and calls static constructors.
+void EmscriptenGlueGenerator::generatePostInstantiateFunction() {
+  Builder builder(wasm);
+  Function* post_instantiate =
+    builder.makeFunction(POST_INSTANTIATE, std::vector<NameType>{}, none, {});
+  wasm.addFunction(post_instantiate);
+
+  if (Function* F = generateAssignGOTEntriesFunction()) {
+    // call __assign_got_enties from post_instantiate
+    Expression* call = builder.makeCall(F->name, {}, none);
+    post_instantiate->body = builder.blockify(post_instantiate->body, call);
+  }
+
+  // The names of standard imports/exports used by lld doesn't quite match that
+  // expected by emscripten.
+  // TODO(sbc): Unify these
+  if (auto* e = wasm.getExportOrNull(WASM_CALL_CTORS)) {
+    Expression* call = builder.makeCall(e->value, {}, none);
+    post_instantiate->body = builder.blockify(post_instantiate->body, call);
+    wasm.removeExport(WASM_CALL_CTORS);
+  }
+
+  auto* ex = new Export();
+  ex->value = post_instantiate->name;
+  ex->name = POST_INSTANTIATE;
+  ex->kind = ExternalKind::Function;
+  wasm.addExport(ex);
+}
+
 Function* EmscriptenGlueGenerator::generateMemoryGrowthFunction() {
   Name name(GROW_WASM_MEMORY);
   std::vector<NameType> params { { NEW_SIZE, i32 } };
@@ -152,12 +311,12 @@ Function* EmscriptenGlueGenerator::generateMemoryGrowthFunction() {
   return growFunction;
 }
 
-static bool hasI64ResultOrParam(FunctionType* ft) {
-  if (ft->result == i64) return true;
-  for (auto ty : ft->params) {
-    if (ty == i64) return true;
-  }
-  return false;
+void EmscriptenGlueGenerator::generateStackInitialization(Address addr) {
+  auto* stackPointer = getStackPointerGlobal();
+  assert(!stackPointer->imported());
+  if (!stackPointer->init || !stackPointer->init->is<Const>())
+    Fatal() << "stack pointer global is not assignable";
+  stackPointer->init->cast<Const>()->value = Literal(int32_t(addr));
 }
 
 inline void exportFunction(Module& wasm, Name name, bool must_export) {
@@ -180,18 +339,20 @@ void EmscriptenGlueGenerator::generateDynCallThunks() {
     tableSegmentData = wasm.table.segments[0].data;
   }
   for (const auto& indirectFunc : tableSegmentData) {
-    if (indirectFunc == dummyFunction) {
-      continue;
-    }
     std::string sig = getSig(wasm.getFunction(indirectFunc));
     auto* funcType = ensureFunctionType(sig, &wasm);
-    if (hasI64ResultOrParam(funcType)) continue; // Can't export i64s on the web.
-    if (!sigs.insert(sig).second) continue; // Sig is already in the set
+    if (!sigs.insert(sig).second) {
+      continue; // sig is already in the set
+    }
+    Name name = std::string("dynCall_") + sig;
+    if (wasm.getFunctionOrNull(name) || wasm.getExportOrNull(name)) {
+      continue; // module already contains this dyncall
+    }
     std::vector<NameType> params;
     params.emplace_back("fptr", i32); // function pointer param
     int p = 0;
     for (const auto& ty : funcType->params) params.emplace_back(std::to_string(p++), ty);
-    Function* f = builder.makeFunction(std::string("dynCall_") + sig, std::move(params), funcType->result, {});
+    Function* f = builder.makeFunction(name, std::move(params), funcType->result, {});
     Expression* fptr = builder.makeGetLocal(0, i32);
     std::vector<Expression*> args;
     for (unsigned i = 0; i < funcType->params.size(); ++i) {
@@ -205,112 +366,64 @@ void EmscriptenGlueGenerator::generateDynCallThunks() {
   }
 }
 
-struct JSCallWalker : public PostWalker<JSCallWalker> {
-  Module &wasm;
-  JSCallWalker(Module &_wasm) : wasm(_wasm) {
-    if (wasm.table.segments.size() == 0) {
-      auto emptySegment =
-          wasm.allocator.alloc<Const>()->set(Literal(uint32_t(0)));
-      wasm.table.segments.emplace_back(emptySegment);
-    }
-    const auto& tableSegmentData = wasm.table.segments[0].data;
+struct RemoveStackPointer : public PostWalker<RemoveStackPointer> {
+  RemoveStackPointer(Global* stackPointer) : stackPointer(stackPointer) {}
 
-    jsCallStartIndex =
-        wasm.table.segments[0].offset->cast<Const>()->value.getInteger();
-    // Check if jsCalls have already been created
-    for (Index i = 0; i < tableSegmentData.size(); ++i) {
-      if (tableSegmentData[i].startsWith("jsCall_")) {
-        jsCallStartIndex += i;
-        return;
-      }
-    }
-    jsCallStartIndex += tableSegmentData.size();
-  }
-
-  // Gather all function signatures used in call_indirect, because any of them
-  // can be used to call function pointers created by emscripten's addFunction.
-  void visitCallIndirect(CallIndirect *curr) {
-    // dynCall thunks are generated in binaryen and call_indirect instructions
-    // within them cannot be used to call function pointers returned by
-    // emscripten's addFunction.
-    if (!getFunction()->name.startsWith("dynCall_")) {
-      indirectlyCallableSigs.insert(
-          getSig(wasm.getFunctionType(curr->fullType)));
+  void visitGetGlobal(GetGlobal* curr) {
+    if (getModule()->getGlobalOrNull(curr->name) == stackPointer) {
+      needStackSave = true;
+      if (!builder) builder = make_unique<Builder>(*getModule());
+      replaceCurrent(builder->makeCall(STACK_SAVE, {}, i32));
     }
   }
 
-  bool createJSCallThunks;
-  Index jsCallStartIndex;
-  // Function type signatures used in call_indirect instructions
-  std::set<std::string> indirectlyCallableSigs;
+  void visitSetGlobal(SetGlobal* curr) {
+    if (getModule()->getGlobalOrNull(curr->name) == stackPointer) {
+      needStackRestore = true;
+      if (!builder) builder = make_unique<Builder>(*getModule());
+      replaceCurrent(builder->makeCall(STACK_RESTORE, {curr->value}, none));
+    }
+  }
+
+  bool needStackSave = false;
+  bool needStackRestore = false;
+
+private:
+  std::unique_ptr<Builder> builder;
+  Global* stackPointer;
 };
 
-JSCallWalker getJSCallWalker(Module& wasm) {
-  JSCallWalker walker(wasm);
-  walker.walkModule(&wasm);
-  return walker;
-}
-
-void EmscriptenGlueGenerator::generateJSCallThunks(
-    unsigned numReservedFunctionPointers) {
-  if (numReservedFunctionPointers == 0)
+void EmscriptenGlueGenerator::replaceStackPointerGlobal() {
+  Global* stackPointer = getStackPointerGlobal();
+  if (!stackPointer)
     return;
 
-  JSCallWalker walker = getJSCallWalker(wasm);
-  auto& tableSegmentData = wasm.table.segments[0].data;
-  for (std::string sig : walker.indirectlyCallableSigs) {
-    // Add imports for jsCall_sig (e.g. jsCall_vi).
-    // Imported jsCall_sig functions have their first parameter as an index to
-    // the function table, so we should prepend an 'i' to parameters' signature
-    // (e.g. If the signature of the callee is 'vi', the imported jsCall_vi
-    // function would have signature 'vii'.)
-    std::string importSig = std::string(1, sig[0]) + 'i' + sig.substr(1);
-    FunctionType *importType = ensureFunctionType(importSig, &wasm);
-    auto import = new Function;
-    import->name = import->base = "jsCall_" + sig;
-    import->module = ENV;
-    import->type = importType->name;
-    FunctionTypeUtils::fillFunction(import, importType);
-    wasm.addFunction(import);
-    FunctionType *funcType = ensureFunctionType(sig, &wasm);
-
-    // Create jsCall_sig_index thunks (e.g. jsCall_vi_0, jsCall_vi_1, ...)
-    // e.g. If # of reserved function pointers (given by a command line
-    // argument) is 3 and there are two possible signature 'vi' and 'ii', the
-    // genereated thunks will be jsCall_vi_0, jsCall_vi_1, jsCall_vi_2,
-    // jsCall_ii_0, jsCall_ii_1, and jsCall_ii_2.
-    for (unsigned fp = 0; fp < numReservedFunctionPointers; ++fp) {
-      std::vector<NameType> params;
-      int p = 0;
-      for (const auto& ty : funcType->params) {
-        params.emplace_back(std::to_string(p++), ty);
-      }
-      Function* f = builder.makeFunction(
-          std::string("jsCall_") + sig + "_" + std::to_string(fp),
-          std::move(params), funcType->result, {});
-      std::vector<Expression*> args;
-      args.push_back(builder.makeConst(Literal(fp)));
-      for (unsigned i = 0; i < funcType->params.size(); ++i) {
-        args.push_back(builder.makeGetLocal(i, funcType->params[i]));
-      }
-      Expression* call =
-          builder.makeCall(import->name, args, funcType->result);
-      f->body = call;
-      wasm.addFunction(f);
-      tableSegmentData.push_back(f->name);
-    }
+  // Replace all uses of stack pointer global
+  RemoveStackPointer walker(stackPointer);
+  walker.walkModule(&wasm);
+  if (walker.needStackSave) {
+    ensureFunctionImport(&wasm, STACK_SAVE, "i");
   }
-  wasm.table.initial = wasm.table.max =
-      wasm.table.segments[0].offset->cast<Const>()->value.getInteger() +
-      tableSegmentData.size();
+  if (walker.needStackRestore) {
+    ensureFunctionImport(&wasm, STACK_RESTORE, "vi");
+  }
+
+  // Finally remove the stack pointer global itself. This avoids importing
+  // a mutable global.
+  wasm.removeGlobal(stackPointer->name);
 }
 
 std::vector<Address> getSegmentOffsets(Module& wasm) {
   std::vector<Address> segmentOffsets;
   for (unsigned i = 0; i < wasm.memory.segments.size(); ++i) {
-    Const* addrConst = wasm.memory.segments[i].offset->cast<Const>();
-    auto address = addrConst->value.geti32();
-    segmentOffsets.push_back(address);
+    if (auto* addrConst = wasm.memory.segments[i].offset->dynCast<Const>()) {
+      auto address = addrConst->value.geti32();
+      segmentOffsets.push_back(address);
+    } else {
+      // TODO(sbc): Wasm shared libraries have data segments with non-const
+      // offset.
+      segmentOffsets.push_back(0);
+    }
   }
   return segmentOffsets;
 }
@@ -363,23 +476,29 @@ std::string codeForConstAddr(Module& wasm,
   return escape(str);
 }
 
-struct AsmConstWalker : public PostWalker<AsmConstWalker> {
+struct AsmConstWalker : public LinearExecutionWalker<AsmConstWalker> {
   Module& wasm;
   std::vector<Address> segmentOffsets; // segment index => address offset
 
   std::map<std::string, std::set<std::string>> sigsForCode;
   std::map<std::string, Address> ids;
   std::set<std::string> allSigs;
+  std::map<Index, SetLocal*> sets; // last sets in the current basic block, per index
 
   AsmConstWalker(Module& _wasm)
     : wasm(_wasm),
       segmentOffsets(getSegmentOffsets(wasm)) { }
 
+  void noteNonLinear(Expression* curr);
+
+  void visitSetLocal(SetLocal* curr);
   void visitCall(Call* curr);
+  void visitTable(Table* curr);
 
   void process();
 
 private:
+  std::string fixupNameWithSig(Name& name, std::string baseSig);
   Literal idLiteralForCode(std::string code);
   std::string asmConstSig(std::string baseSig);
   Name nameForImportWithSig(std::string sig);
@@ -389,21 +508,65 @@ private:
   std::vector<std::unique_ptr<Function>> queuedImports;
 };
 
+void AsmConstWalker::noteNonLinear(Expression* curr) {
+  // End of this basic block; clear sets.
+  sets.clear();
+}
+
+void AsmConstWalker::visitSetLocal(SetLocal* curr) {
+  sets[curr->index] = curr;
+}
+
 void AsmConstWalker::visitCall(Call* curr) {
   auto* import = wasm.getFunction(curr->target);
+  // Find calls to emscripten_asm_const* functions whose first argument is
+  // is always a string constant.
   if (import->imported() && import->base.hasSubstring(EMSCRIPTEN_ASM_CONST)) {
-    auto arg = curr->operands[0]->cast<Const>();
-    auto code = codeForConstAddr(wasm, segmentOffsets, arg);
-    arg->value = idLiteralForCode(code);
     auto baseSig = getSig(curr);
-    auto sig = asmConstSig(baseSig);
-    sigsForCode[code].insert(sig);
-    auto importName = nameForImportWithSig(sig);
-    curr->target = importName;
+    auto sig = fixupNameWithSig(curr->target, baseSig);
+    auto* arg = curr->operands[0];
+    while (!arg->dynCast<Const>()) {
+      if (auto* get = arg->dynCast<GetLocal>()) {
+        // The argument may be a local.get, in which case, the last set in this
+        // basic block has the value.
+        auto* set = sets[get->index];
+        if (set) {
+          assert(set->index == get->index);
+          arg = set->value;
+        }
+      } else if (auto* value = arg->dynCast<Binary>()) {
+        // In the dynamic linking case the address of the string constant
+        // is the result of adding its offset to __memory_base.
+        // In this case are only looking for the offset with the data segment so
+        // the RHS of the addition is just what we want.
+        assert(value->op == AddInt32);
+        arg = value->right;
+      } else {
+        if (!value) {
+          Fatal() << "Unexpected arg0 type (" << getExpressionName(arg)
+              << ") in call to to: " << import->base;
+        }
+      }
+    }
 
-    if (allSigs.count(sig) == 0) {
-      allSigs.insert(sig);
-      queueImport(importName, baseSig);
+    auto* value = arg->cast<Const>();
+    auto code = codeForConstAddr(wasm, segmentOffsets, value);
+    sigsForCode[code].insert(sig);
+
+    // Replace the first argument to the call with a Const index
+    Builder builder(wasm);
+    curr->operands[0] = builder.makeConst(idLiteralForCode(code));
+  }
+}
+
+void AsmConstWalker::visitTable(Table* curr) {
+  for (auto& segment : curr->segments) {
+    for (auto& name : segment.data) {
+      auto* func = wasm.getFunction(name);
+      if (func->imported() && func->base.hasSubstring(EMSCRIPTEN_ASM_CONST)) {
+        std::string baseSig = getSig(func);
+        fixupNameWithSig(name, baseSig);
+      }
     }
   }
 }
@@ -416,6 +579,18 @@ void AsmConstWalker::process() {
   addImports();
 }
 
+std::string AsmConstWalker::fixupNameWithSig(Name& name, std::string baseSig) {
+  auto sig = asmConstSig(baseSig);
+  auto importName = nameForImportWithSig(sig);
+  name = importName;
+
+  if (allSigs.count(sig) == 0) {
+    allSigs.insert(sig);
+    queueImport(importName, baseSig);
+  }
+  return sig;
+}
+
 Literal AsmConstWalker::idLiteralForCode(std::string code) {
   int32_t id;
   if (ids.count(code) == 0) {
@@ -499,16 +674,31 @@ struct EmJsWalker : public PostWalker<EmJsWalker> {
     auto addrConst = curr->body->dynCast<Const>();
     if (addrConst == nullptr) {
       auto block = curr->body->dynCast<Block>();
-      Expression* first = nullptr;
+      Expression* value = nullptr;
       if (block && block->list.size() > 0) {
-        first = block->list[0];
+        value = block->list[0];
+        // first item may be a set of a local that we get later
+        auto* set = value->dynCast<SetLocal>();
+        if (set) {
+          value = block->list[1];
+        }
+        // look into a return value
+        if (auto* ret = value->dynCast<Return>()) {
+          value = ret->value;
+        }
+        // if it's a get of that set, use that value
+        if (auto* get = value->dynCast<GetLocal>()) {
+          if (set && get->index == set->index) {
+            value = set->value;
+          }
+        }
       }
-      if (first) {
-        addrConst = first->dynCast<Const>();
+      if (value) {
+        addrConst = value->dynCast<Const>();
       }
     }
     if (addrConst == nullptr) {
-      Fatal() << "Unexpected generated __em_js__ function body: " << curr;
+      Fatal() << "Unexpected generated __em_js__ function body: " << curr->name;
     }
     auto code = codeForConstAddr(wasm, segmentOffsets, addrConst);
     codeByName[funcName] = code;
@@ -532,12 +722,6 @@ EmJsWalker fixEmJsFuncsAndReturnWalker(Module& wasm) {
   return walker;
 }
 
-void EmscriptenGlueGenerator::fixEmAsmConsts() {
-  fixEmAsmConstsAndReturnWalker(wasm);
-  fixEmJsFuncsAndReturnWalker(wasm);
-}
-
-
 // Fixes function name hacks caused by LLVM exception & setjmp/longjmp
 // handling pass for wasm.
 // This does two things:
@@ -630,31 +814,11 @@ struct FixInvokeFunctionNamesWalker : public PostWalker<FixInvokeFunctionNamesWa
     }
   }
 
-  void visitTable(Table* curr) {
-    for (auto& segment : curr->segments) {
-      for (size_t i = 0; i < segment.data.size(); i++) {
-        auto it = importRenames.find(segment.data[i]);
-        if (it != importRenames.end()) {
-          segment.data[i] = it->second;
-        }
-      }
-    }
-  }
-
-  void visitCall(Call* curr) {
-    if (wasm.getFunction(curr->target)->imported()) {
-      auto it = importRenames.find(curr->target);
-      if (it != importRenames.end()) {
-        curr->target = it->second;
-      }
-    }
-  }
-
   void visitModule(Module* curr) {
     for (auto importName : toRemove) {
       wasm.removeFunction(importName);
     }
-    wasm.updateMaps();
+    ModuleUtils::renameFunctions(wasm, importRenames);
   }
 };
 
@@ -676,70 +840,67 @@ void printSet(std::ostream& o, C& c) {
 }
 
 std::string EmscriptenGlueGenerator::generateEmscriptenMetadata(
-    Address staticBump, std::vector<Name> const& initializerFunctions,
-    unsigned numReservedFunctionPointers) {
+    Address staticBump, std::vector<Name> const& initializerFunctions) {
   bool commaFirst;
-  auto maybeComma = [&commaFirst]() {
+  auto nextElement = [&commaFirst]() {
     if (commaFirst) {
       commaFirst = false;
-      return "";
+      return "\n    ";
     } else {
-      return ",";
+      return ",\n    ";
     }
   };
 
   std::stringstream meta;
-  meta << "{ ";
+  meta << "{\n";
 
   AsmConstWalker emAsmWalker = fixEmAsmConstsAndReturnWalker(wasm);
 
   // print
   commaFirst = true;
-  meta << "\"asmConsts\": {";
-  for (auto& pair : emAsmWalker.sigsForCode) {
-    auto& code = pair.first;
-    auto& sigs = pair.second;
-    meta << maybeComma();
-    meta << '"' << emAsmWalker.ids[code] << "\": [\"" << code << "\", ";
-    printSet(meta, sigs);
-    meta << "]";
+  if (!emAsmWalker.sigsForCode.empty()) {
+    meta << "  \"asmConsts\": {";
+    for (auto& pair : emAsmWalker.sigsForCode) {
+      auto& code = pair.first;
+      auto& sigs = pair.second;
+      meta << nextElement();
+      meta << '"' << emAsmWalker.ids[code] << "\": [\"" << code << "\", ";
+      printSet(meta, sigs);
+      meta << ", ";
+
+      // TODO: proxying to main thread. Currently this is unsupported, so proxy
+      // mode is "none", represented by an empty string.
+      meta << "[\"\"]";
+
+      meta << "]";
+    }
+    meta << "\n  },\n";
   }
-  meta << "},";
 
   EmJsWalker emJsWalker = fixEmJsFuncsAndReturnWalker(wasm);
-  if (emJsWalker.codeByName.size() > 0) {
-    meta << "\"emJsFuncs\": {";
+  if (!emJsWalker.codeByName.empty()) {
+    meta << "  \"emJsFuncs\": {";
     commaFirst = true;
     for (auto& pair : emJsWalker.codeByName) {
       auto& name = pair.first;
       auto& code = pair.second;
-      meta << maybeComma();
+      meta << nextElement();
       meta << '"' << name << "\": \"" << code << '"';
     }
-    meta << "},";
+    meta << "\n  },\n";
   }
 
-  meta << "\"staticBump\": " << staticBump << ", ";
+  meta << "  \"staticBump\": " << staticBump << ",\n";
+  meta << "  \"tableSize\": " << wasm.table.initial.addr << ",\n";
 
-  meta << "\"initializers\": [";
-  commaFirst = true;
-  for (const auto& func : initializerFunctions) {
-    meta << maybeComma();
-    meta << "\"" << func.c_str() << "\"";
-  }
-  meta << "]";
-
-  if (numReservedFunctionPointers) {
-    JSCallWalker jsCallWalker = getJSCallWalker(wasm);
-    meta << ", ";
-    meta << "\"jsCallStartIndex\": " << jsCallWalker.jsCallStartIndex << ", ";
-    meta << "\"jsCallFuncType\": [";
+  if (!initializerFunctions.empty()) {
+    meta << "  \"initializers\": [";
     commaFirst = true;
-    for (std::string sig : jsCallWalker.indirectlyCallableSigs) {
-      meta << maybeComma();
-      meta << "\"" << sig << "\"";
+    for (const auto& func : initializerFunctions) {
+      meta << nextElement();
+      meta << "\"" << func.c_str() << "\"";
     }
-    meta << "]";
+    meta << "\n  ],\n";
   }
 
   // Avoid adding duplicate imports to `declares' or `invokeFuncs`.  Even
@@ -751,57 +912,99 @@ std::string EmscriptenGlueGenerator::generateEmscriptenMetadata(
   // We use the `base` rather than the `name` of the imports here and below
   // becasue this is the externally visible name that the embedder (JS) will
   // see.
-  meta << ", \"declares\": [";
+  meta << "  \"declares\": [";
   commaFirst = true;
   ModuleUtils::iterImportedFunctions(wasm, [&](Function* import) {
     if (emJsWalker.codeByName.count(import->base.str) == 0 &&
         !import->base.startsWith(EMSCRIPTEN_ASM_CONST.str) &&
-        !import->base.startsWith("invoke_") &&
-        !import->base.startsWith("jsCall_")) {
+        !import->base.startsWith("invoke_")) {
       if (declares.insert(import->base.str).second) {
-        meta << maybeComma() << '"' << import->base.str << '"';
+        meta << nextElement() << '"' << import->base.str << '"';
       }
     }
   });
-  meta << "]";
+  meta << "\n  ],\n";
 
-  meta << ", \"externs\": [";
+  meta << "  \"externs\": [";
   commaFirst = true;
   ModuleUtils::iterImportedGlobals(wasm, [&](Global* import) {
-    meta << maybeComma() << "\"_" << import->base.str << '"';
-  });
-  meta << "]";
-
-  meta << ", \"implementedFunctions\": [";
-  commaFirst = true;
-  for (const auto& ex : wasm.exports) {
-    if (ex->kind == ExternalKind::Function) {
-      meta << maybeComma() << "\"_" << ex->name.str << '"';
+    if (!(import->module == ENV && import->name == STACK_INIT)) {
+      meta << nextElement() << "\"_" << import->base.str << '"';
     }
-  }
-  meta << "]";
+  });
+  meta << "\n  ],\n";
 
-  meta << ", \"exports\": [";
-  commaFirst = true;
-  for (const auto& ex : wasm.exports) {
-    meta << maybeComma() << '"' << ex->name.str << '"';
-  }
-  meta << "]";
+  if (!wasm.exports.empty()) {
+    meta << "  \"implementedFunctions\": [";
+    commaFirst = true;
+    for (const auto& ex : wasm.exports) {
+      if (ex->kind == ExternalKind::Function) {
+        meta << nextElement() << "\"_" << ex->name.str << '"';
+      }
+    }
+    meta << "\n  ],\n";
 
-  meta << ", \"invokeFuncs\": [";
+    meta << "  \"exports\": [";
+    commaFirst = true;
+    for (const auto& ex : wasm.exports) {
+      if (ex->kind == ExternalKind::Function) {
+        meta << nextElement() << '"' << ex->name.str << '"';
+      }
+    }
+    meta << "\n  ],\n";
+
+    meta << "  \"namedGlobals\": {";
+    commaFirst = true;
+    for (const auto& ex : wasm.exports) {
+      if (ex->kind == ExternalKind::Global) {
+        const Global* g = wasm.getGlobal(ex->value);
+        assert(g->type == i32);
+        Const* init = g->init->cast<Const>();
+        uint32_t addr = init->value.geti32();
+        meta << nextElement() << '"' << ex->name.str << "\" : \"" << addr << '"';
+      }
+    }
+    meta << "\n  },\n";
+  }
+
+  meta << "  \"invokeFuncs\": [";
   commaFirst = true;
   ModuleUtils::iterImportedFunctions(wasm, [&](Function* import) {
     if (import->base.startsWith("invoke_")) {
       if (invokeFuncs.insert(import->base.str).second) {
-        meta << maybeComma() << '"' << import->base.str << '"';
+        meta << nextElement() << '"' << import->base.str << '"';
       }
     }
   });
-  meta << "]";
+  meta << "\n  ],\n";
 
-  meta << " }\n";
+  meta << "  \"features\": [";
+  commaFirst = true;
+  wasm.features.iterFeatures([&](FeatureSet::Feature f) {
+    meta << nextElement() << "\"--enable-" << FeatureSet::toString(f) << '"';
+  });
+  meta << "\n  ]\n";
+
+  meta << "}\n";
 
   return meta.str();
 }
 
+void EmscriptenGlueGenerator::separateDataSegments(Output* outfile, Address base) {
+  size_t lastEnd = 0;
+  for (Memory::Segment& seg : wasm.memory.segments) {
+    assert(!seg.isPassive && "separating passive segments not implemented");
+    size_t offset = seg.offset->cast<Const>()->value.geti32();
+    offset -= base;
+    size_t fill = offset - lastEnd;
+    if (fill > 0) {
+      std::vector<char> buf(fill);
+      outfile->write(buf.data(), fill);
+    }
+    outfile->write(seg.data.data(), seg.data.size());
+    lastEnd = offset + seg.data.size();
+  }
+  wasm.memory.segments.clear();
+}
+
 } // namespace wasm
